// Unit.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import java.io.*;

import swarm.gui.Raster;
import swarm.gui.ZoomRasterImpl;

/**
 * The Unit class instances are micro unit in our virtual enterprise;
 * i.e. the units where the steps required to build a product (to fulfill
 * an order) are done
 *
 * @author Pietro Terna (with Marco Remondino, Dario Landini, Nicola 
 * Ormezzano and Paolo Pelligra contributions)
 */
public class Unit extends SwarmObjectImpl {

    /** the number identifying the unit*/
    public int unitNumber;
    /** the unit production phase */
    public int unitProductionPhase;

    /** the list of the orders to be executed */
    public ListImpl  waitingList;
    /** the list containing the orders (if any) performed in a day */
    public ListImpl dailyProductionList;
    /** the modelSwarm (to which we apply the 
     * setProductionTimeAndRecipeLength(...) method */
    ESFrameModelSwarm eSFrameModelSwarm;
    /** the temporary address of an existing unit */
    public Unit aUnit;
    /** the first order to be executed */
    public Order firstOrder;
    /** position of the unit into the display */
    public int xPos, yPos; 
   /** position of the unit into the display */
    public byte unitColor;
    /** the unitActivitySpace */
    public UnitActivitySpace unitActivitySpace;
    /** our visibility */
    public double visibility;
    /** increasing visibility in each step 1 */
    public double increasingVisibilityStep;
    /** the number of tick of inactivity */
    public int inactivity;
    /** unsent product at the end of step2 */
    public int unsentProducts;

    /** the instance of AssigningTool we use */
    AssigningTool assigningTool; 

    /**
     * the constructor for Unit
     */
    public Unit (Zone aZone, ESFrameModelSwarm mo, UnitActivitySpace uas, 
		 AssigningTool at, int x, int y, byte col, double vis, double visStep) {

	// Call the constructor for the Unit parent class
	super(aZone);

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the model
	eSFrameModelSwarm = mo;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());
	// the assigning tool
	assigningTool=at;
	// x and y pos
	xPos=x;
	yPos=y;
	// the color
	unitColor=col;
	// the unit activity space
	unitActivitySpace = uas;
	// our visibility
	visibility = vis;
	// increasing visibility
	increasingVisibilityStep=visStep;
	// initial number of inativity ticks
	inactivity=0;
	// initial value of unsent products
	unsentProducts=0;

   }

    /** set unit number */
    public void setUnitNumber(int un)
    {
	// the unit id number.
	unitNumber=un;
	// announce the unit to the console
	if(StartESFrame.verbose)
	System.out.println("Unit number " + unitNumber + " has been created.");
     }

    /** get unit number */
    public int getUnitNumber()
    {
	// the unit id number.
	return unitNumber;
    }


    /**
     * production:<br><br>
     *
     * at the beginning of each time tick in the 'day', each unit looks for 
     * the order(s) (if any) to be executed; the production 
     * cycle is 'one unit of time'
     * in a step of the recipe of an order in a tick of a day;<br><br>
     */
    public void unitStep1 ()
    {
	int i, ii; 

	if(StartESFrame.verbose)
	    System.out.println("Step 1 in unit " + unitNumber +
			       "; waitingList count " + 
			       waitingList.getCount());

	// showing the visibility
	unitActivitySpace.showVisibility(xPos, yPos, (int) visibility);

	// increasing visibility
	visibility += increasingVisibilityStep;

	// inactivity
	if(waitingList.getCount() == 0) inactivity++;
	unitActivitySpace.showDownHistogram(xPos, yPos, inactivity);

	// the production

	// after the specific step in recipe production we place the order(s) 
	// in a daily production
	// list, which will be used at unitStep2 to diffuse production
	// results

	if(waitingList.getCount() > 0){
	    inactivity=0;
	    firstOrder=(Order)waitingList.removeFirst();

	    if(firstOrder.getNextStep()!=getProductionPhase()){
		System.out.println("Unrecoverable error (case2) in unit " 
				   + unitNumber + " (order number "
				   + firstOrder.getOrderNumber()+ " step "
				   + firstOrder.getNextStep()+")");
		MyExit.exit(1);}
	

	    firstOrder.setDoneStep(unitNumber);
	    dailyProductionList.addLast(firstOrder);
	    eSFrameModelSwarm.addToGlobalProduction(1);
	}

    } // concluding step1


    /**
     * diffusion and accounting:<br><br>
     *
     * sending the orders to the subsequent production units;<br><br>
     *
     * making accounting operations
     */
    public void unitStep2 () {
	float timeSpent;
	int numberOfProd, product;
	Order notAssignedOrder;

	// the leftHistogram related to the waitingList
	unitActivitySpace.showLeftHistogram(xPos, yPos, getWaitingListLength());

	if(StartESFrame.verbose)
	    System.out.println("Step 2 in unit " + unitNumber +
			       " dailyProductionList count " +
			       dailyProductionList.getCount());

	numberOfProd=dailyProductionList.getCount();
	for (product=0;product<numberOfProd;product++)
	    {
		firstOrder=(Order)dailyProductionList.removeFirst();
		
		// do we have concluded the production step?
		// if yes ... we drop the object
		if (firstOrder.getNextStep()==0)
		    {
			if(StartESFrame.verbose)
			    System.out.println("drop order # " +
					       firstOrder.getOrderNumber() +
					       " startingTimeUnit " +
					       firstOrder.getStartingTimeUnit() +
					       " endTimeUnit " +
					       Globals.env.getCurrentTime() +
					       " stepNumber " +
					       firstOrder.getStepNumber() +
					       " ratio " +
					       ( (float)
						 (Globals.env.getCurrentTime() -
						  firstOrder.getStartingTimeUnit()))/
					       firstOrder.getStepNumber());
			// the production is concluded, so we inform
			// our environment (the model) of the time spent
			// and of the # of steps in recipe
			timeSpent=Globals.env.getCurrentTime() -
			    firstOrder.getStartingTimeUnit();
			eSFrameModelSwarm.setProductionTimeAndRecipeLength(
					 timeSpent,
					 firstOrder.getNumberOfDirectStepsDone());
			if(StartESFrame.verbose)
			    System.out.println("Order #" +
					       firstOrder.orderNumber +
					       " has been completed; time spent in ticks: " +
					       timeSpent +
					       " expected time in ticks " +
					       firstOrder.getStepNumber() + 
					       " or with stand alone batches " + 
					       firstOrder.getNumberOfDirectStepsDone());

			// adding the production to the concluded order value
			eSFrameModelSwarm.addToConcludedOrders(firstOrder.
							       getNumberOfDirectStepsDone());
			if(StartESFrame.verbose)
			    System.out.println("concluded " + 
					       firstOrder.getNumberOfDirectStepsDone());
			firstOrder.drop();
		    }
		
		// if not ... we send it to the subsequent production unit
		// if non found, the order is placed into the dailyProductionList
		else 
		    {
			notAssignedOrder=assigningTool.assign(this, firstOrder);
			if (notAssignedOrder != null)dailyProductionList.
							 addLast(notAssignedOrder);
		    }
	    }

	// the rightHistogram related to the dailyProductionList (here the value is
	// that un unsent products)
	unsentProducts = dailyProductionList.getCount();
	unitActivitySpace.showRightHistogram(xPos, yPos, unsentProducts);

    } // concluding step2

    /** set unit production phases */
    public void setProductionPhase(int ph)
    {
	unitProductionPhase = ph;
    }


    /**
     * return the production phases of the unit
     */
    public int getProductionPhase()
    {
	return unitProductionPhase;
    }

    /**
     * Putting an order in the production waitingList of the unit
     * (using a related object if any)
     */
    public void setWaitingList(Order anOrder)
    {
	int i; 

	// normal case: 
	waitingList.addLast(anOrder);

	if(StartESFrame.verbose)
	    System.out.println("The unit " + unitNumber +
			       " has received the order # "
			       + anOrder.getOrderNumber());
    }

    /** the visibility */
    public int getVisibility()
    {
	return (int) visibility;
    }

    /** the inactivity */
    public int getInactivity()
    {
	return inactivity;
    }

    /** the X position */
    public int getXPos()
    {
	return xPos;
    }

    /** the Y position */
    public int getYPos()
    {
	return yPos;
    }

    /** the length of the wainting list */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }
    


    /** unsent products at the end of step 2 */
    public int getUnsentProducts()
    {
	return unsentProducts;
    }

    /** get daily production list */
    public ListImpl getDailyProductionList(){
	return dailyProductionList;
    }

    /** get waiting list */
    public ListImpl getWaitingList(){
	return waitingList;
    }

    /** designing itself on the raster */
    public Object drawSelfOn (Raster r)
    {
	r.drawPointX$Y$Color (xPos, yPos, unitColor);
	return this;
    }

    /** getting the model the unit belongs to */
    public Object getUnitModel()
    {
	return eSFrameModelSwarm;
    }

}
